Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq tuple.
Require Import fintype finfun bigop ssralg poly polydiv.
Require Import zmodp matrix mxpoly polyXY fieldext.
Require Import div prime binomial choice generic_quotient.

(******************************************************************************)
(* This file is supposed to provide theory of separable and inseparable field *)
(* extensions, however it is currently half about general field extentions    *)
(* and this half should perhaps be move elsewhere.                            *)
(*                                                                            *)
(*  separablePolynomial p == p has no repeated roots in any field extension   *)
(*   separableElement K x == the minimal polynomial for x is separable        *)
(*          separable K E == every member of E is separable over K            *)
(* separableGenerator K E == some x \in E that generates the largest possible *)
(*                           subfield K[x] that is separable over K           *)
(* purelyInseparableElement K x == there is n \in [char L].-nat such that     *)
(*                                 x ^+ n \in K                               *)
(*  purelyInseparable K E == every member of E is purely inseparable over K   *)
(*                                                                            *)
(*  Derivations are only meant as a tool to prove allSeparableElement         *)
(*         Derivation K D == D is a linear operator on K that satifies        *)
(*                           Leibniz's product rule                           *)
(* DerivationExtend x D K == Given a derivation D on K and a separable        *)
(*                           element x over K, this function returns the      *)
(*                           unique extension of D to K(x).                   *)
(******************************************************************************)


Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Open Local Scope ring_scope.

Import GRing.Theory.

(* :TODO: Move to polydiv.v *)
Lemma coprimep_addl_mul (R : idomainType) (p q r : {poly R}) :
  coprimep r (p * r + q) = coprimep r q.
Proof. by rewrite !coprimep_def (eqp_size (gcdp_addl_mul _ _ _)). Qed.

Lemma exprp_addl (R : comRingType) (x y : R) (n : nat) :
  [char R].-nat n -> (x + y) ^+ n = x ^+ n + y ^+ n.
Proof.
case: n=> [|[|]] // n Hn; move: (Hn)=> /pnatPpi /(_ (pi_pdiv _))=> charR.
move: Hn; rewrite (eq_pnat _ (charf_eq charR))=> /p_natP [] e ->.
elim: e => // e ihe; rewrite !expnSr !exprn_mulr ihe.
by rewrite [_ ^+ _]Frobenius_aut_add_comm //; apply: mulrC.
Qed.

Section SeparablePoly.

Variable (R: idomainType).
Implicit Types (p q d u v : {poly R}).

Definition separablePolynomial p := coprimep p p^`().

Local Notation sep := separablePolynomial.
Local Notation lcn_neq0 := (ID.lc_expn_scalp_neq0 _).

Lemma separable_neq0 p : separablePolynomial p -> p != 0.
Proof. by apply: contraL=> /eqP ->; rewrite /sep deriv0 coprime0p eqp01. Qed.

Lemma nosquareP p :
  (forall u v, u * v %| p -> coprimep u v)
  <-> (forall u, size u != 1%N -> ~~ (u ^+ 2 %| p)).
Proof.
split=> [hp u pu_neq1|hp u v dvd_uv_p].
  by apply/negP=> /hp; rewrite coprimepp; apply/negP.
rewrite coprimep_def; have [//|/hp/negPf<-] := altP eqP.
by rewrite (dvdp_trans _ dvd_uv_p) // dvdp_mul ?dvdp_gcdl ?dvdp_gcdr.
Qed.

Lemma separablePolynomialP p :
  reflect ((forall u v, u * v %| p -> coprimep u v)
        /\ (forall u, u %| p -> 1 < size u -> u^`() != 0))
          (sep p).
Proof.
apply: (iffP idP)=> [sep_p|].
  split; last first.
    move=> u dvd_up u_gt1; apply/negP=> /eqP u_eq0.
    suff: u %| gcdp p p^`().
      by rewrite gtNdvdp -?size_poly_eq0 // coprimep_size_gcd.
    pose c := lead_coef u ^+ scalp p u.
    have c_neq0 : c != 0 by rewrite lcn_neq0.
    rewrite dvdp_gcd dvd_up -(@dvdp_scaler _ c) // -derivZ -ID.divpK //=.
    by rewrite derivM u_eq0 mulr0 addr0 dvdp_mull.
  move=> u v; rewrite ID.dvdp_eq; set c := _ ^+ _.
  have c_neq0 : c != 0 by rewrite ID.lc_expn_scalp_neq0.
  rewrite mulrA; set r := _ * u => /eqP p_eq; move: sep_p; rewrite /sep.
  rewrite -(coprimep_scalel _ _ c_neq0) -(coprimep_scaler _ _ c_neq0).
  rewrite -derivZ p_eq derivM coprimep_mull.
  rewrite {1}addrC mulrC !coprimep_addl_mul !(coprimep_mull, coprimep_mulr).
  by do ![case/andP|move=> ?].
move=> [hp2 hp1]; rewrite /sep coprimep_def; set g := gcdp _ _.
pose c := lead_coef g ^+ scalp p g.
have c_neq0 : c != 0 by rewrite lcn_neq0.
have [|p_gt1|//] := ltngtP.
  rewrite ltnS leqn0 size_poly_eq0 /g gcdp_eq0=> /andP[/eqP p_eq0 _].
  by have := hp2 0 0; rewrite mulr0 p_eq0 dvdp0 coprimep0 eqp01; apply.
have: g %| (c *: p^`()) by rewrite dvdp_scaler ?dvdp_gcdr.
rewrite -derivZ -ID.divpK ?dvdp_gcdl // derivM.
rewrite dvdp_addr; last by rewrite dvdp_mull.
rewrite gausspr ?hp2 1?mulrC ?ID.divpK ?dvdp_gcdl -/c ?dvdp_scalel //.
move=> /dvdp_leq; rewrite leqNgt lt_size_deriv; last first.
  by rewrite -size_poly_gt0 (leq_trans _ p_gt1).
by apply; rewrite hp1 ?dvdp_gcdl.
Qed.

Lemma separable_coprime p : sep p -> forall u v, u * v %| p -> coprimep u v.
Proof. by move=> /separablePolynomialP[]. Qed.

Lemma separable_nosquare p : sep p ->
  forall u k, 1%N < k -> u %| p -> size u != 1%N -> u ^+ k %| p = false.
Proof.
move=> /separablePolynomialP[/nosquareP hp _] u [|[|]] // k _ hu su.
have [|//] := boolP (_ %| _) => /(dvdp_trans _) => /(_ (u ^+ 2)).
by rewrite (negPf (hp _ _)) // => -> //; rewrite expr2 !exprS mulrA dvdp_mulr.
Qed.

Lemma separable_deriv_eq0 p : sep p ->
  forall u, u %| p -> 1 < size u -> (u^`() == 0) = false.
Proof.
by move=> /separablePolynomialP[_ h] u hu su; apply: negbTE; rewrite h.
Qed.

(* Cyril : Could it be useful ? *)
(* Lemma decompose_dvdp p q : coprimep p q -> *)
(*   forall u, u %| p * q -> u %= gcdp u p * gcdp u q. *)
(* Proof. *)
(* move=> cpq u dvd_u_pq; rewrite /eqp gaussp_inv ?dvdp_gcdl ?andbT; last first. *)
(*   by rewrite (coprimep_dvdl (dvdp_gcdr _ _)) ?(coprimep_dvdr (dvdp_gcdr _ _)). *)
(* have [|pq_neq0] := boolP ((p * q) == 0). *)
(*   by rewrite mulf_eq0=> /orP[] /eqP->; rewrite gcdp0 (dvdp_mulIr, dvdp_mulIl). *)
(* rewrite -(@dvdp_mul2l _ (p * q %/ u)) ?dvdp_div_eq // ID.divpK //. *)
(* rewrite dvdp_scalel ?lcn_neq0 ?gaussp_inv // mulrA [in X in _ && X]mulrAC. *)
(* rewrite !(eqp_dvdr _ (eqp_mulr _ (mulp_gcdr _ _ _))) ?ID.divpK //. *)
(* rewrite !(eqp_dvdr _ (eqp_mulr _ (gcdp_scalel _ _ _))) ?lcn_neq0 // dvdp_mulr. *)
(*   by rewrite dvdp_mulr // (eqp_dvdr _ (gcdp_mul2r _ _ _)) dvdp_mull. *)
(* by rewrite mulrC (eqp_dvdr _ (gcdp_mul2r _ _ _)) dvdp_mull. *)
(* Qed. *)

Lemma dvdp_separable p q : q %| p -> sep p -> sep q.
Proof.
move=> dvd_qp sep_p; apply/separablePolynomialP; split=> [u v huv|u hu su].
  by rewrite (separable_coprime sep_p) // (dvdp_trans huv).
by rewrite (separable_deriv_eq0 sep_p) // (dvdp_trans hu).
Qed.

Lemma separable_mul p q : separablePolynomial (p * q) =
 [&& separablePolynomial p, separablePolynomial q & coprimep p q].
Proof.
apply/idP/and3P => [H|].
  by rewrite !(dvdp_separable _ H) (dvdp_mulIr,dvdp_mulIl,separable_coprime H).
rewrite /sep=> [] [Hp Hq Hpq]; rewrite derivM coprimep_mull {1}addrC mulrC.
rewrite !coprimep_addl_mul !coprimep_def !(eqp_size (gaussp_gcdr _ _)) //.
by rewrite -?coprimep_def Hpq coprimep_sym.
Qed.

Lemma eqp_separable p q : p %= q -> sep p = sep q.
Proof. by move=> /andP[hpq hqp]; apply/idP/idP=> /dvdp_separable; apply. Qed.

(* (* Cyril : I liked these Def, Hint and two lemmas *) *)
(* Definition irredp p := forall d, size d != 1%N -> d %| p -> d %= p. *)

(* Lemma irredp_factor x : irredp ('X - x%:P). *)
(* Proof. *)
(* move=> d size_d dvd_df; rewrite -dvdp_size_eqp // size_factor. *)
(* have := dvdp_leq _ dvd_df; rewrite factor_eq0 size_factor=> /(_ isT). *)
(* case hs: size size_d dvd_df=> [|[|[]]] // _. *)
(* by move/eqP: hs; rewrite size_poly_eq0=> /eqP->; rewrite dvd0p factor_eq0. *)
(* Qed. *)
(* Hint Resolve irredp_factor. *)

(* Lemma irredp_factorP d p : irredp p -> d %| p -> {d %= 1} + {d %= p}. *)
(* Proof. *)
(* move=> irred_p dvd_dp; have [] := boolP (_ %= 1); first by left. *)
(* by rewrite -size_poly_eq1=> /irred_p /(_ dvd_dp); right. *)
(* Qed. *)

(* :TODO: change polydiv.v lemma to this *)
Lemma irredp_factor (x : R) (d : {poly R}) :
   d %| 'X - x%:P -> (d %= 1) || (d %= 'X - x%:P).
Proof. by move=> /irredp_factor; rewrite size_poly_eq1. Qed.

Lemma separable_root p x : sep (p * ('X - x%:P)) = sep p && ~~ root p x.
Proof.
rewrite separable_mul; have [sep_p /=|//] := boolP (sep p).
rewrite /sep deriv_factor coprimep1 rootE coprimep_sym /=.
apply/idP/idP; first by move=> /coprimep_root/=/(_ x) ->; rewrite ?root_factor.
move=> px_neq0; apply/coprimepP=> d /irredp_factor /orP[] // eq_d.
by rewrite (eqp_dvdl _ eq_d) dvdp_factorl rootE (negPf px_neq0).
Qed.

Lemma separable_factors (r : seq R) :
  separablePolynomial (\prod_(x <- r) ('X - x%:P)) = uniq r.
Proof.
elim: r => [|x r IH]; first by rewrite big_nil /separablePolynomial coprime1p.
by rewrite big_cons mulrC separable_root IH root_prod_factors andbC.
Qed.

Lemma make_separable p :  p != 0 -> sep (p %/ (gcdp p p^`())).
Proof.
move=> p_neq0; set g := gcdp _ _; apply/separablePolynomialP.
have max_dvd_u : forall u : {poly R}, 1 < size u -> exists k, ~~ (u ^+ k %| p).
  move=> u size_u; exists (size p); rewrite gtNdvdp // [X in _ < X]polySpred.
    by rewrite size_exp_id ltnS leq_pmull //; case: size size_u.
  by rewrite expf_neq0 // -size_poly_gt0 (leq_trans _ size_u).
split; last first.
  move=> u hu size_u; apply: contraL isT=> /eqP u'_eq0 /=.
  have [//|[|k]] := ex_minnP (max_dvd_u u _); first by rewrite dvd1p.
  move=> hk /(_ k); rewrite ltnn; apply; apply: contra hk=> hk.
  suff: u ^+ k.+1 %| (p %/ g) * g.
    by rewrite ID.divpK ?dvdp_gcdl // dvdp_scaler ?lcn_neq0.
  rewrite exprS dvdp_mul // dvdp_gcd hk //=.
  suff: u ^+ k %| ((p %/ u ^+ k) * u ^+ k)^`().
    by rewrite ID.divpK ?dvdp_gcdl // derivZ dvdp_scaler ?lcn_neq0.
  by rewrite ?dvdp_gcdl // !derivCE u'_eq0 mul0r mul0rn mulr0 addr0 dvdp_mull.
apply/nosquareP=> u; have [|size_u _|//] := ltngtP.
  rewrite // ltnS leqn0 size_poly_eq0=> /eqP -> _; rewrite exprS mul0r.
  apply/negP=> /dvdp_trans/=/(_ _ (divp_dvd _)).
  by rewrite dvd0p (negPf p_neq0) dvdp_gcdl=> /implyP.
have [hu|//] := boolP (_ %| _); have [//|] := ex_minnP (max_dvd_u u _).
move=> [|[|n hn]]; first by rewrite dvd1p.
  rewrite expr1 (@dvdp_trans _ (p %/ g)) ?divp_dvd ?dvdp_gcdl //.
  by rewrite (dvdp_trans _ hu) // exprS dvdp_mulr.
move=> /(_ n.+1); rewrite ltnn; apply; apply: contra hn=> hn.
suff: u ^+ n.+2 %| (p %/ g) * g.
  by rewrite ID.divpK ?dvdp_gcdl // dvdp_scaler ?lcn_neq0.
rewrite !exprS mulrA dvdp_mul //.
rewrite dvdp_gcd (dvdp_trans _ hn) ?exprS ?dvdp_mull //=.
suff: u ^+ n %| ((p %/ u ^+ n.+1) * u ^+ n.+1)^`().
  by rewrite ID.divpK ?dvdp_gcdl // derivZ dvdp_scaler ?lcn_neq0.
by rewrite !derivCE dvdp_add // -1?mulr_natl ?exprS !dvdp_mull.
Qed.

End SeparablePoly.

(* :TODO: Move this to polydiv.v *)
Lemma egcdp_rec_map (F : fieldType) (R : idomainType)
  (f : {rmorphism F -> R}) (p q : {poly F}) n :
    (map_poly f (egcdp_rec p q n).1, map_poly f (egcdp_rec p q n).2) =
    (egcdp_rec (map_poly f p) (map_poly f q) n).
Proof.
elim: n p q => [|n IH] => /= p q; first by rewrite rmorph1 rmorph0.
rewrite map_poly_eq0.
case: eqP => Hq0; first by rewrite rmorph1 rmorph0.
rewrite -map_modp -(IH q (p %% q)).
case: (egcdp_rec _ _ n) => a b /=.
rewrite map_poly_scaler lead_coef_map -rmorphX scalp_map rmorph_sub rmorphM.
by rewrite -map_divp.
Qed.

(* :TODO: Move this to polydiv.v *)
Lemma egcdp_map (F : fieldType) (R : idomainType)
  (f : {rmorphism F -> R}) (p q : {poly F}) :
    (map_poly f (egcdp p q).1, map_poly f (egcdp p q).2) =
    (egcdp (map_poly f p) (map_poly f q)).
Proof.
rewrite /egcdp !size_map_poly.
case: ifP=> /= hspq; first by apply: egcdp_rec_map.
move: (egcdp_rec_map f q p (size p)).
by case: (egcdp_rec (map_poly _ _) _ _)=> [a b [-> ->]].
Qed.

Section InfinitePrimitiveElementTheorem.

Local Notation "p ^ f" := (map_poly f p) : ring_scope.
Local Notation "'Y" := 'X%:P : ring_scope.

Variables (F L: fieldType) (iota : {rmorphism F -> L}).
Variables (x y : L) (p : {poly F}).
Hypothesis (pne0 : p != 0).
Hypothesis (Hpx : root (p ^ iota) x).

Lemma PET_Infinite_Case : forall q : {poly F},
  root (q ^ iota) y -> separablePolynomial q ->
  exists r : {poly L}, r != 0 /\
   forall t, ~~root r (iota t) ->
    (exists p0, (p0 ^ iota).[iota t * y - x] = x) /\
    (exists q0, (q0 ^ iota).[iota t * y - x] = y).
Proof.
move => q Hqy Hsep.
have qne0 := separable_neq0 Hsep.
set p' := (p ^ iota) \Po ('X + x%:P).
have [qq Hqq] := (factor_theorem _ _ Hqy).
set q' := qq \Po ('X + y%:P).
move: Hsep.
rewrite /separablePolynomial -gcdp_eqp1 -(eqp_map iota) rmorph1 gcdp_map.
rewrite -deriv_map Hqq gcdp_eqp1 [coprimep _ _]separable_root.
case/andP => _.
rewrite /root {1}(_ : y = ('X + y%:P).[0]); last by rewrite !horner_lin.
rewrite -horner_comp_poly -/q'.
have p'ne0 : p' != 0.
 move: pne0.
 apply: contra.
 move/eqP/(f_equal (fun q => q \Po ('X - x%:P)))/eqP.
 rewrite comp_poly0p /p' comp_polyA comp_poly_translateK comp_polyX.
 by rewrite map_poly_eq0.
move=> q'0_neq0; have q'ne0 : q' != 0.
  by apply: contra q'0_neq0=> /eqP->; rewrite horner0.
have : coprimep ((p' ^ polyC) \Po ('Y * 'X)) (q' ^ polyC).
  rewrite coprimep_def -[_ == _]negbK neq_ltn ltnS size_poly_leq0.
  rewrite gcdp_eq0 mulrC poly_XmY_eq0 map_polyC_eq0.
  rewrite (negPf p'ne0) (negPf q'ne0) /= -resultant_eq0.
  by rewrite annul_div_neq0.
rewrite -gcdp_eqp1.
move: (bezoutp (p' ^ polyC \Po 'Y * 'X) (q' ^ polyC)) => [[u v] /= Huv].
rewrite -(eqp_ltrans Huv) -size_poly_eq1.
case/size1P => {Huv} r Hr0 Hr.
exists r.
split => // t Ht.
suff Hq0 : (exists q0 : {poly F}, (q0 ^ iota).[iota t * y - x] = y).
 split => //.
 case: Hq0 => q0 Hq0.
 exists (t *: q0 - 'X).
 rewrite rmorph_sub [_ 'X]map_polyX [_ (_ *: _)]map_poly_scaler !horner_lin.
 by rewrite Hq0 oppr_sub addrC addrNK.
have Hcomm: (commr_rmorph idfun (iota t)) by apply: mulrC.
move/(f_equal (map_poly (horner_morph Hcomm))) : Hr.
rewrite rmorphD !{1}rmorphM map_polyC /= /comp_poly -horner_map rmorphM.
rewrite [_ 'Y]map_polyC [_ 'X]horner_morphX [_ 'X]map_polyX.
rewrite -map_comp_poly ?rmorph0 // -[(q' ^ _) ^ _]map_comp_poly ?rmorph0 //.
rewrite ![GRing.Additive.apply _]/=.
rewrite [_ ^ (map_poly _ \o _)]map_polyE.
rewrite (_ : (map (map_poly (horner_morph Hcomm) \o polyC) (p' ^ polyC))
           = (map (polyC \o (horner_morph Hcomm)) (p' ^ polyC))); last first.
 by rewrite (eq_map (map_polyC _)).
rewrite -map_polyE -map_comp_poly ?rmorph0 //.
rewrite -[((polyC \o horner_morph Hcomm) \o polyC)]
        /(polyC \o (horner_morph Hcomm \o polyC)).
rewrite [p' ^ _]map_comp_poly ?rmorph0 // ![GRing.Additive.apply _]/=.
rewrite !map_polyE !(eq_map (horner_morphC _)) !map_id -!map_polyE.
rewrite !polyseqK -/(comp_poly ((iota t)%:P  * 'X) p').
set u1 := (u ^ _).
set v1 := (v ^ _).
set p1 := (_ \Po _).
rewrite /horner_morph map_polyE map_id polyseqK.
move => Hlincomb.
have : (coprimep p1 q').
 apply/bezout_coprimepP.
 exists (u1, v1).
 by rewrite Hlincomb polyC_eqp1.
clear -Hpx Hqy pne0 qne0 Hqq.
move/(coprimep_comp_poly ('X - y%:P)).
rewrite /q' [(qq \Po _) \Po _]comp_polyA comp_poly_translateK comp_polyX.
rewrite -gcdp_eqp1.
set p2 := (_ \Po _) => Hp2.
have: (gcdp p2 (q ^ iota) %= ('X - y%:P)).
 apply/andP; split; last first.
  rewrite -root_factor_theorem root_gcd Hqy andbT /root /p2 /p1 /p'.
  by rewrite !(horner_comp_poly, horner_lin) subrr mulr0 add0r.
 rewrite -[_ - _]mul1r.
 apply: (@dvdp_trans _ (gcdp (p2 * ('X - y%:P)) (q ^ iota))).
  rewrite dvdp_gcd dvdp_gcdr (dvdp_trans (dvdp_gcdl _ _)) // dvdp_mulr //.
 case/andP: Hp2.
 rewrite Hqq -(eqp_dvdl _ (mulp_gcdl _ _ _)) dvdp_mul2r //.
 by rewrite -size_poly_eq0 size_factor.
set z := iota t * y - x.
have [qt qtne0 Hqt] : exists2 qt, qt != 0 & root (qt ^ iota) (iota t * y).
 case (eqVneq t 0) => [-> | Ht].
  exists 'X; first by rewrite -size_poly_eq0 size_polyX.
  by rewrite rmorph0 mul0r /root map_polyX hornerX.
 exists (q \Po (t^-1 *: 'X)).
  move: qne0.
  apply: contra.
  move/eqP/polyP => Hqt.
  apply/eqP/polyP => i.
  rewrite coef0 -[X in _ = X](mulr0 (t ^+ i)) -[X in _ * X](coef0 _ i) -(Hqt i).
  rewrite -coefZ comp_polyE scaler_sumr -{1}[q]coefK poly_def !coef_sum.
  apply: eq_bigr => j _.
  rewrite scaler_exp !coefZ mulrA [t ^+ i * _]mulrC -mulrA coefXn.
  case: (eqVneq i j) => [-> | Hij].
   by rewrite expr_inv mulVKf // expf_eq0 negb_and Ht orbT.
  by rewrite -[_ == _]negbK Hij !mulr0.
 apply/eqP.
 move/eqP: Hqy <-.
 rewrite comp_polyE rmorph_sum /root horner_sum horner_coef size_map_poly.
 apply eq_bigr => i _ /=.
 rewrite map_poly_scaler rmorphX /= map_poly_scaler map_polyX coef_map.
 rewrite horner_scaler horner_exp horner_scaler hornerX.
 by rewrite fmorphV mulKf // fmorph_eq0.
pose f := annul_sub qt p.
have Hf0 : f != 0 by rewrite annul_sub_neq0.
have Hfz : root (f ^ iota) z by rewrite rootE annul_sub_iotaP //; apply/rootP.
set Fz := subFExtend iota z f.
set kappa := (@subfx_inj _ _ _ _ _) : Fz -> L.
pose (Q := (q ^ (inj_subfx iota z f))).
have HQ : q ^ iota = Q ^ kappa.
 rewrite /Q -map_comp_poly ?[kappa 0]rmorph0 // !map_polyE.
 congr (Poly _); apply: eq_map => a.
 by rewrite /= /kappa subfx_inj_eval // map_polyC hornerC.
pose (P := (p ^ (inj_subfx iota z f) \Po
  ((inj_subfx iota z f t) *: 'X - (subfx_eval iota z f 'X)%:P))).
have HP : p2 = P ^ kappa.
 rewrite -!horner_map rmorph_sub /= map_poly_scaler map_polyC map_polyX /=.
 rewrite !subfx_inj_eval // map_polyX map_polyC hornerX hornerC.
 rewrite -map_comp_poly ?rmorph0 //.
 rewrite (eq_map_poly (map_polyC _)).
 rewrite map_comp_poly ?rmorph0 //=.
 rewrite -[(p ^ _) ^ _]map_comp_poly ?rmorph0 //.
 rewrite (eq_map_poly (fun q => subfx_inj_eval (polyC q) Hf0 Hfz)).
 rewrite (eq_map_poly (horner_morphC _)); last by move => ?; apply: mulrC.
 rewrite /p2 /p1 /p' !comp_polyA {2 3}/comp_poly.
 rewrite rmorphM rmorphD ![GRing.RMorphism.apply _]/= !map_polyX !map_polyC.
 rewrite !horner_lin mulr_addr -addrA -polyC_opp -polyC_mul polyC_add oppr_add.
 by rewrite -!polyC_opp opprK mulrN mul_polyC.
rewrite HQ HP -gcdp_map /=.
move/eqpP => [[c1 c2]] /andP [Hc1 Hc2].
move/(canLR (scalerK Hc2)).
rewrite scalerA.
move/polyP => Hgcd.
move:(Hgcd 1%N).
rewrite coefD coefN coefC coefX subr0 coefZ coef_map mulr1n.
move => Hgcd1.
move:(nonzero1r L).
rewrite -Hgcd1 mulf_eq0 negb_or.
move/andP => [Hc120 Hgcd10].
move/eqP:(Hgcd 0%N).
rewrite coefD coefN coefC coefX coefZ coef_map sub0r -eqr_oppC.
move:Hgcd1.
move/(canRL (mulfK Hgcd10)) ->.
rewrite mul1r -fmorphV -rmorphM -rmorphN.
set y' := -_.
case: (subfxE y') => [h ->].
rewrite /= subfx_inj_eval //.
move/eqP => Hh.
by exists h.
Qed.

Lemma PET_char0 : forall q : {poly F},
  q != 0 -> root (q ^ iota) y -> [char F] =i pred0 -> exists n,
  (exists p0, (p0 ^ iota).[y *+ n - x] = x) /\
  (exists q0, (q0 ^ iota).[y *+ n - x] = y).
Proof.
move => q qne0 Hqy; move/charf0P => Hchar.
move: (dvdp_gcdl q q^`()).
(* assia : should use dvdpP here *)
rewrite dvdp_eq.
set qq := _ %/ _ => /eqP Hq.
have Hqqy : root (qq ^ iota) y.
  move: (qne0); rewrite -(map_poly_eq0 iota).
  case/(maxdivp y) => r Hry [[|n] Hqr].
    by  move: Hqr Hqy Hry ->; rewrite expr0 mulr1 => ->.
 move: (f_equal (@deriv _) Hqr).
 rewrite deriv_map derivM deriv_exp deriv_factor mul1r exprS -mulr_natl !mulrA.
 rewrite -mulr_addl => Hq'.
 have: ('X - y%:P) ^+ n.+1 %| q ^iota by rewrite Hqr; apply: dvdp_mulIr.
 move: (f_equal (map_poly iota) Hq).
 rewrite rmorphM /=.
 rewrite gcdp_map {2}Hqr.
 rewrite Hq' exprS mulrA => ->.
 set s1 := r * _.
 set s2 := r^`() * _ + _.
 set s3 := _ ^+ n.
 move: (mulp_gcdl s1 s2 s3).
 move:qne0.
 rewrite Hq mulf_eq0 negb_or -(map_poly_eq0 iota).
 case/andP.
 move/eqp_mul2l => <- _.
 move/eqp_dvdr <-.
 rewrite mulrA dvdp_mul2r /s3 ?expf_eq0 ?factor_eq0 ?andbF //.
 rewrite -root_factor_theorem root_mul root_gcd.
 case/orP => //.
 case/andP => _.
 rewrite root_factor_theorem dvdp_addr; last by rewrite dvdp_mull // dvdpp.
 rewrite -root_factor_theorem mulr_natr -scaler_nat root_scaler; last first.
  rewrite -(rmorph0 iota) -(rmorph1 iota) -rmorphMn (inj_eq (fmorph_inj _)).
  by rewrite Hchar.
 by rewrite -[_ r y]negbK Hry.
have Hsep: separablePolynomial qq.
 move: (make_separable qne0).
 rewrite {1}Hq mulpK; last by rewrite gcdp_eq0 negb_and qne0.
 by move/dvdp_separable; apply.
case: (PET_Infinite_Case Hqqy Hsep) => f [Hf0 Hf].
pose s := mkseq (fun x => iota (x%:R : F)) (size f).
have Hs : uniq_roots s.
 rewrite uniq_rootsE.
 apply: mkseq_uniq.
 suff Hwlog: forall a b, iota (a%:R : F) = iota (b%:R : F) -> a <= b;
  move => a b Hab.
  apply/eqP.
  rewrite eqn_leq.
  rewrite !Hwlog //.
 move/eqP: Hab.
 apply: contraLR.
 rewrite -ltnNge => Hab.
 rewrite (inj_eq (fmorph_inj iota)) -subr_eq0 -mulrn_subr; last by rewrite ltnW.
 move: Hab.
 rewrite Hchar -subn_gt0.
 by move/prednK <-.
move/contra: (fun X => max_ring_poly_roots (rs:=s) Hf0 X Hs).
rewrite -leqNgt size_mkseq.
move/(_ (leqnn _)).
case/allPn => ?.
case/mapP => t _ -> Ht.
exists t.
rewrite -mulr_natl -[1](rmorph1 iota) -rmorphMn.
by apply: Hf.
Qed.

End InfinitePrimitiveElementTheorem.
